@@ -13,20 +13,25 @@ def parse_variable_spec(spec):
 def select_matrix_environments(environments, included_variables, excluded_variables):
     selected_environments = []
     for env_name, variables in environments.items():
+        included = set(variables)
+        excluded = set()
+
         for variable, value in variables.items():
             if variable in excluded_variables:
                 excluded_values = excluded_variables[variable]
                 if not excluded_values or value in excluded_values:
+                    excluded.add(variable)
                     break
 
             if included_variables:
                 if variable not in included_variables:
-                    break
+                    included.remove(variable)
                 else:
                     included_values = included_variables[variable]
                     if included_values and value not in included_values:
-                        break
-        else:
+                        included.remove(variable)
+
+        if included and not excluded:
             selected_environments.append(env_name)
 
     return selected_environments
@@ -74,34 +79,25 @@ def run(
     select or exclude certain variables, optionally followed by specific comma-separated values.
     For example, if you have the following configuration:
 
-    === ":octicons-file-code-16: pyproject.toml"
-
-        \b
-        ```toml
-        [[tool.hatch.envs.test.matrix]]
-        python = ["39", "310"]
-        version = ["42", "3.14", "9000"]
-        ```
-
-    === ":octicons-file-code-16: hatch.toml"
-
-        \b
-        ```toml
-        [[envs.test.matrix]]
-        python = ["39", "310"]
-        version = ["42", "3.14", "9000"]
-        ```
+    \b
+    ```toml config-example
+    [[tool.hatch.envs.test.matrix]]
+    python = ["3.9", "3.10"]
+    version = ["42", "3.14", "9000"]
+    ```
 
     then running:
 
     \b
     ```
-    hatch env run -i py=310 -x version=9000 test:pytest
+    hatch env run -i py=3.10 -x version=9000 test:pytest
     ```
 
-    would execute `pytest` in the environments `test.py310-42` and `test.py310-3.14`.
+    would execute `pytest` in the environments `test.py3.10-42` and `test.py3.10-3.14`.
     Note that `py` may be used as an alias for `python`.
     """
+    app.ensure_environment_plugin_dependencies()
+
     project = app.project
 
     included_variables = {}
@@ -167,12 +163,12 @@ def run(
 
             try:
                 environment.check_compatibility()
-            except Exception as e:
+            except Exception as e:  # noqa: BLE001
                 if ignore_compat or matrix_selected:
                     incompatible[environment.name] = str(e)
                     continue
-                else:
-                    app.abort(f'Environment `{env_name}` is incompatible: {e}')
+
+                app.abort(f'Environment `{env_name}` is incompatible: {e}')
 
             any_compatible = True
             if should_display_header:
