@@ -5,31 +5,31 @@ import sys
 from abc import ABC, abstractmethod
 from contextlib import contextmanager
 from os.path import isabs
+from typing import TYPE_CHECKING
 
 from hatch.config.constants import AppEnvVars
 from hatch.env.utils import add_verbosity_flag
 from hatch.project.utils import format_script_commands, parse_script_command
 from hatch.utils.structures import EnvVars
 
+if TYPE_CHECKING:
+    from collections.abc import Iterable
+
 
 class EnvironmentInterface(ABC):
     """
     Example usage:
 
-    === ":octicons-file-code-16: plugin.py"
-
-        ```python
+    ```python tab="plugin.py"
         from hatch.env.plugin.interface import EnvironmentInterface
 
 
         class SpecialEnvironment(EnvironmentInterface):
             PLUGIN_NAME = 'special'
             ...
-        ```
+    ```
 
-    === ":octicons-file-code-16: hooks.py"
-
-        ```python
+    ```python tab="hooks.py"
         from hatchling.plugin import hookimpl
 
         from .plugin import SpecialEnvironment
@@ -38,19 +38,32 @@ class EnvironmentInterface(ABC):
         @hookimpl
         def hatch_register_environment():
             return SpecialEnvironment
-        ```
+    ```
     """
 
     PLUGIN_NAME = ''
     """The name used for selection."""
 
-    def __init__(self, root, metadata, name, config, matrix_variables, data_directory, platform, verbosity, app=None):
+    def __init__(
+        self,
+        root,
+        metadata,
+        name,
+        config,
+        matrix_variables,
+        data_directory,
+        isolated_data_directory,
+        platform,
+        verbosity,
+        app=None,
+    ):
         self.__root = root
         self.__metadata = metadata
         self.__name = name
         self.__config = config
         self.__matrix_variables = matrix_variables
         self.__data_directory = data_directory
+        self.__isolated_data_directory = isolated_data_directory
         self.__platform = platform
         self.__verbosity = verbosity
         self.__app = app
@@ -128,24 +141,26 @@ class EnvironmentInterface(ABC):
     @property
     def data_directory(self):
         """
-        The [directory](../../config/hatch.md#environments) reserved exclusively for this plugin as a path-like object.
+        The [directory](../../config/hatch.md#environments) this plugin should use for storage as a path-like object.
+        If the user has not configured one then this will be the same as the
+        [isolated data directory](reference.md#hatch.env.plugin.interface.EnvironmentInterface.isolated_data_directory).
         """
         return self.__data_directory
 
+    @property
+    def isolated_data_directory(self):
+        """
+        The default [directory](../../config/hatch.md#environments) reserved exclusively for this plugin as a path-like
+        object.
+        """
+        return self.__isolated_data_directory
+
     @property
     def config(self) -> dict:
         """
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>]
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>]
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>]
+        ```
         """
         return self.__config
 
@@ -172,28 +187,22 @@ class EnvironmentInterface(ABC):
     @property
     def env_vars(self) -> dict:
         """
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>.env-vars]
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>.env-vars]
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>.env-vars]
+        ```
         """
         if self._env_vars is None:
             env_vars = self.config.get('env-vars', {})
             if not isinstance(env_vars, dict):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.env-vars` must be a mapping')
+                message = f'Field `tool.hatch.envs.{self.name}.env-vars` must be a mapping'
+                raise TypeError(message)
 
             for key, value in env_vars.items():
                 if not isinstance(value, str):
-                    raise TypeError(
+                    message = (
                         f'Environment variable `{key}` of field `tool.hatch.envs.{self.name}.env-vars` must be a string'
                     )
+                    raise TypeError(message)
 
             new_env_vars = {}
             with self.metadata.context.apply_context(self.context):
@@ -208,28 +217,21 @@ class EnvironmentInterface(ABC):
     @property
     def env_include(self) -> list[str]:
         """
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>]
-            env-include = [...]
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>]
-            env-include = [...]
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>]
+        env-include = [...]
+        ```
         """
         if self._env_include is None:
             env_include = self.config.get('env-include', [])
             if not isinstance(env_include, list):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.env-include` must be an array')
+                message = f'Field `tool.hatch.envs.{self.name}.env-include` must be an array'
+                raise TypeError(message)
 
             for i, pattern in enumerate(env_include, 1):
                 if not isinstance(pattern, str):
-                    raise TypeError(f'Pattern #{i} of field `tool.hatch.envs.{self.name}.env-include` must be a string')
+                    message = f'Pattern #{i} of field `tool.hatch.envs.{self.name}.env-include` must be a string'
+                    raise TypeError(message)
 
             if env_include:
                 self._env_include = ['HATCH_BUILD_*', *env_include]
@@ -241,28 +243,21 @@ class EnvironmentInterface(ABC):
     @property
     def env_exclude(self) -> list[str]:
         """
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>]
-            env-exclude = [...]
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>]
-            env-exclude = [...]
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>]
+        env-exclude = [...]
+        ```
         """
         if self._env_exclude is None:
             env_exclude = self.config.get('env-exclude', [])
             if not isinstance(env_exclude, list):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.env-exclude` must be an array')
+                message = f'Field `tool.hatch.envs.{self.name}.env-exclude` must be an array'
+                raise TypeError(message)
 
             for i, pattern in enumerate(env_exclude, 1):
                 if not isinstance(pattern, str):
-                    raise TypeError(f'Pattern #{i} of field `tool.hatch.envs.{self.name}.env-exclude` must be a string')
+                    message = f'Pattern #{i} of field `tool.hatch.envs.{self.name}.env-exclude` must be a string'
+                    raise TypeError(message)
 
             self._env_exclude = env_exclude
 
@@ -274,24 +269,25 @@ class EnvironmentInterface(ABC):
             from packaging.requirements import InvalidRequirement, Requirement
 
             dependencies_complex = []
-            with self.metadata.context.apply_context(self.context):
+            with self.apply_context():
                 for option in ('dependencies', 'extra-dependencies'):
                     dependencies = self.config.get(option, [])
                     if not isinstance(dependencies, list):
-                        raise TypeError(f'Field `tool.hatch.envs.{self.name}.{option}` must be an array')
+                        message = f'Field `tool.hatch.envs.{self.name}.{option}` must be an array'
+                        raise TypeError(message)
 
                     for i, entry in enumerate(dependencies, 1):
                         if not isinstance(entry, str):
-                            raise TypeError(
+                            message = (
                                 f'Dependency #{i} of field `tool.hatch.envs.{self.name}.{option}` must be a string'
                             )
+                            raise TypeError(message)
 
                         try:
                             dependencies_complex.append(Requirement(self.metadata.context.format(entry)))
                         except InvalidRequirement as e:
-                            raise ValueError(
-                                f'Dependency #{i} of field `tool.hatch.envs.{self.name}.{option}` is invalid: {e}'
-                            )
+                            message = f'Dependency #{i} of field `tool.hatch.envs.{self.name}.{option}` is invalid: {e}'
+                            raise ValueError(message) from None
 
             self._environment_dependencies_complex = dependencies_complex
 
@@ -310,16 +306,29 @@ class EnvironmentInterface(ABC):
     @property
     def dependencies_complex(self):
         if self._dependencies_complex is None:
-            dependencies_complex = list(self.environment_dependencies_complex)
+            all_dependencies_complex = list(self.environment_dependencies_complex)
 
             # Ensure these are checked last to speed up initial environment creation since
             # they will already be installed along with the project
-            if not self.skip_install and self.dev_mode:
-                dependencies_complex.extend(self.metadata.core.dependencies_complex.values())
-                for feature in self.features:
-                    dependencies_complex.extend(self.metadata.core.optional_dependencies_complex[feature].values())
+            if (not self.skip_install and self.dev_mode) or self.features:
+                from hatch.utils.dep import get_project_dependencies_complex
 
-            self._dependencies_complex = dependencies_complex
+                dependencies_complex, optional_dependencies_complex = get_project_dependencies_complex(self)
+
+                if not self.skip_install and self.dev_mode:
+                    all_dependencies_complex.extend(dependencies_complex.values())
+
+                for feature in self.features:
+                    if feature not in optional_dependencies_complex:
+                        message = (
+                            f'Feature `{feature}` of field `tool.hatch.envs.{self.name}.features` is not '
+                            f'defined in the dynamic field `project.optional-dependencies`'
+                        )
+                        raise ValueError(message)
+
+                    all_dependencies_complex.extend(optional_dependencies_complex[feature].values())
+
+            self._dependencies_complex = all_dependencies_complex
 
         return self._dependencies_complex
 
@@ -328,7 +337,8 @@ class EnvironmentInterface(ABC):
         """
         The list of all [project dependencies](../../config/metadata.md#dependencies) (if
         [installed](../../config/environment/overview.md#skip-install) and in
-        [dev mode](../../config/environment/overview.md#dev-mode)) and
+        [dev mode](../../config/environment/overview.md#dev-mode)), selected
+        [optional dependencies](../../config/environment/overview.md#features), and
         [environment dependencies](../../config/environment/overview.md#dependencies).
         """
         if self._dependencies is None:
@@ -341,28 +351,21 @@ class EnvironmentInterface(ABC):
         """
         All names are stored as their lower-cased version.
 
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>]
-            platforms = [...]
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>]
-            platforms = [...]
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>]
+        platforms = [...]
+        ```
         """
         if self._platforms is None:
             platforms = self.config.get('platforms', [])
             if not isinstance(platforms, list):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.platforms` must be an array')
+                message = f'Field `tool.hatch.envs.{self.name}.platforms` must be an array'
+                raise TypeError(message)
 
             for i, command in enumerate(platforms, 1):
                 if not isinstance(command, str):
-                    raise TypeError(f'Platform #{i} of field `tool.hatch.envs.{self.name}.platforms` must be a string')
+                    message = f'Platform #{i} of field `tool.hatch.envs.{self.name}.platforms` must be a string'
+                    raise TypeError(message)
 
             self._platforms = [platform.lower() for platform in platforms]
 
@@ -371,24 +374,16 @@ class EnvironmentInterface(ABC):
     @property
     def skip_install(self) -> bool:
         """
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>]
-            skip-install = ...
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>]
-            skip-install = ...
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>]
+        skip-install = ...
+        ```
         """
         if self._skip_install is None:
             skip_install = self.config.get('skip-install', not self.metadata.has_project_file())
             if not isinstance(skip_install, bool):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.skip-install` must be a boolean')
+                message = f'Field `tool.hatch.envs.{self.name}.skip-install` must be a boolean'
+                raise TypeError(message)
 
             self._skip_install = skip_install
 
@@ -397,24 +392,16 @@ class EnvironmentInterface(ABC):
     @property
     def dev_mode(self) -> bool:
         """
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>]
-            dev-mode = ...
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>]
-            dev-mode = ...
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>]
+        dev-mode = ...
+        ```
         """
         if self._dev_mode is None:
             dev_mode = self.config.get('dev-mode', True)
             if not isinstance(dev_mode, bool):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.dev-mode` must be a boolean')
+                message = f'Field `tool.hatch.envs.{self.name}.dev-mode` must be a boolean'
+                raise TypeError(message)
 
             self._dev_mode = dev_mode
 
@@ -427,25 +414,35 @@ class EnvironmentInterface(ABC):
 
             features = self.config.get('features', [])
             if not isinstance(features, list):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.features` must be an array of strings')
+                message = f'Field `tool.hatch.envs.{self.name}.features` must be an array of strings'
+                raise TypeError(message)
 
             all_features = set()
             for i, feature in enumerate(features, 1):
                 if not isinstance(feature, str):
-                    raise TypeError(f'Feature #{i} of field `tool.hatch.envs.{self.name}.features` must be a string')
-                elif not feature:
-                    raise ValueError(
-                        f'Feature #{i} of field `tool.hatch.envs.{self.name}.features` cannot be an empty string'
-                    )
+                    message = f'Feature #{i} of field `tool.hatch.envs.{self.name}.features` must be a string'
+                    raise TypeError(message)
 
-                feature = normalize_project_name(feature)
-                if feature not in self.metadata.core.optional_dependencies:
-                    raise ValueError(
-                        f'Feature `{feature}` of field `tool.hatch.envs.{self.name}.features` is not '
+                if not feature:
+                    message = f'Feature #{i} of field `tool.hatch.envs.{self.name}.features` cannot be an empty string'
+                    raise ValueError(message)
+
+                normalized_feature = (
+                    feature
+                    if self.metadata.hatch.metadata.allow_ambiguous_features
+                    else normalize_project_name(feature)
+                )
+                if (
+                    not self.metadata.hatch.metadata.hook_config
+                    and normalized_feature not in self.metadata.core.optional_dependencies
+                ):
+                    message = (
+                        f'Feature `{normalized_feature}` of field `tool.hatch.envs.{self.name}.features` is not '
                         f'defined in field `project.optional-dependencies`'
                     )
+                    raise ValueError(message)
 
-                all_features.add(feature)
+                all_features.add(normalized_feature)
 
             self._features = sorted(all_features)
 
@@ -454,24 +451,16 @@ class EnvironmentInterface(ABC):
     @property
     def description(self) -> str:
         """
-        === ":octicons-file-code-16: pyproject.toml"
-
-            ```toml
-            [tool.hatch.envs.<ENV_NAME>]
-            description = ...
-            ```
-
-        === ":octicons-file-code-16: hatch.toml"
-
-            ```toml
-            [envs.<ENV_NAME>]
-            description = ...
-            ```
+        ```toml config-example
+        [tool.hatch.envs.<ENV_NAME>]
+        description = ...
+        ```
         """
         if self._description is None:
             description = self.config.get('description', '')
             if not isinstance(description, str):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.description` must be a string')
+                message = f'Field `tool.hatch.envs.{self.name}.description` must be a string'
+                raise TypeError(message)
 
             self._description = description
 
@@ -482,15 +471,17 @@ class EnvironmentInterface(ABC):
         if self._scripts is None:
             script_config = self.config.get('scripts', {})
             if not isinstance(script_config, dict):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.scripts` must be a table')
+                message = f'Field `tool.hatch.envs.{self.name}.scripts` must be a table'
+                raise TypeError(message)
 
             config = {}
 
             for name, data in script_config.items():
                 if ' ' in name:
-                    raise ValueError(
+                    message = (
                         f'Script name `{name}` in field `tool.hatch.envs.{self.name}.scripts` must not contain spaces'
                     )
+                    raise ValueError(message)
 
                 commands = []
 
@@ -499,15 +490,17 @@ class EnvironmentInterface(ABC):
                 elif isinstance(data, list):
                     for i, command in enumerate(data, 1):
                         if not isinstance(command, str):
-                            raise TypeError(
+                            message = (
                                 f'Command #{i} in field `tool.hatch.envs.{self.name}.scripts.{name}` must be a string'
                             )
+                            raise TypeError(message)
 
                         commands.append(command)
                 else:
-                    raise TypeError(
+                    message = (
                         f'Field `tool.hatch.envs.{self.name}.scripts.{name}` must be a string or an array of strings'
                     )
+                    raise TypeError(message)
 
                 config[name] = commands
 
@@ -525,13 +518,15 @@ class EnvironmentInterface(ABC):
         if self._pre_install_commands is None:
             pre_install_commands = self.config.get('pre-install-commands', [])
             if not isinstance(pre_install_commands, list):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.pre-install-commands` must be an array')
+                message = f'Field `tool.hatch.envs.{self.name}.pre-install-commands` must be an array'
+                raise TypeError(message)
 
             for i, command in enumerate(pre_install_commands, 1):
                 if not isinstance(command, str):
-                    raise TypeError(
+                    message = (
                         f'Command #{i} of field `tool.hatch.envs.{self.name}.pre-install-commands` must be a string'
                     )
+                    raise TypeError(message)
 
             self._pre_install_commands = list(pre_install_commands)
 
@@ -542,13 +537,15 @@ class EnvironmentInterface(ABC):
         if self._post_install_commands is None:
             post_install_commands = self.config.get('post-install-commands', [])
             if not isinstance(post_install_commands, list):
-                raise TypeError(f'Field `tool.hatch.envs.{self.name}.post-install-commands` must be an array')
+                message = f'Field `tool.hatch.envs.{self.name}.post-install-commands` must be an array'
+                raise TypeError(message)
 
             for i, command in enumerate(post_install_commands, 1):
                 if not isinstance(command, str):
-                    raise TypeError(
+                    message = (
                         f'Command #{i} of field `tool.hatch.envs.{self.name}.post-install-commands` must be a string'
                     )
+                    raise TypeError(message)
 
             self._post_install_commands = list(post_install_commands)
 
@@ -598,6 +595,10 @@ class EnvironmentInterface(ABC):
         This should perform the necessary steps to completely remove the environment from the system and will only
         be triggered manually by users with the [`env remove`](../../cli/reference.md#hatch-env-remove) or
         [`env prune`](../../cli/reference.md#hatch-env-prune) commands.
+
+        If the
+        [build environment](reference.md#hatch.env.plugin.interface.EnvironmentInterface.build_environment)
+        has a caching mechanism, this should remove that as well.
         """
 
     @abstractmethod
@@ -644,8 +645,25 @@ class EnvironmentInterface(ABC):
         in the environment.
         """
 
+    def dependency_hash(self):
+        """
+        This should return a hash of the environment's
+        [dependencies](reference.md#hatch.env.plugin.interface.EnvironmentInterface.dependencies)
+        and any other data that is handled by the
+        [sync_dependencies](reference.md#hatch.env.plugin.interface.EnvironmentInterface.sync_dependencies)
+        and
+        [dependencies_in_sync](reference.md#hatch.env.plugin.interface.EnvironmentInterface.dependencies_in_sync)
+        methods.
+        """
+        from hatch.utils.dep import hash_dependencies
+
+        return hash_dependencies(self.dependencies_complex)
+
     @contextmanager
-    def build_environment(self, dependencies: list[str]):
+    def build_environment(
+        self,
+        dependencies: list[str],  # noqa: ARG002
+    ):
         """
         This should set up an isolated environment in which to [`build`](../../cli/reference.md#hatch-build) the project
         given a set of dependencies and must be a context manager:
@@ -663,27 +681,44 @@ class EnvironmentInterface(ABC):
         with self.get_env_vars():
             yield
 
-    def get_build_process(self, build_environment, **kwargs):
+    def run_builder(
+        self,
+        build_environment,  # noqa: ARG002
+        **kwargs,
+    ):
         """
         This will be called when the
         [build environment](reference.md#hatch.env.plugin.interface.EnvironmentInterface.build_environment)
         is active:
 
         ```python
-        with environment.build_environment([...]) as build_environment:
-            build_process = environment.get_build_process(build_environment, ...)
+        with environment.build_environment([...]) as build_env:
+            process = environment.run_builder(build_env, ...)
         ```
 
         This should return the standard library's
-        [subprocess.Popen](https://docs.python.org/3/library/subprocess.html#subprocess.Popen)
-        with all output captured by `stdout`. The command is constructed by passing all keyword arguments to
+        [subprocess.CompletedProcess](https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess).
+        The command is constructed by passing all keyword arguments to
         [construct_build_command](reference.md#hatch.env.plugin.interface.EnvironmentInterface.construct_build_command).
 
         For an example, open the default implementation below:
         """
-        return self.platform.capture_process(self.construct_build_command(**kwargs))
+        return self.platform.run_command(self.construct_build_command(**kwargs))
 
-    def enter_shell(self, name, path, args):
+    def build_environment_exists(self):  # noqa: PLR6301
+        """
+        If the
+        [build environment](reference.md#hatch.env.plugin.interface.EnvironmentInterface.build_environment)
+        has a caching mechanism, this should indicate whether or not it has already been created.
+        """
+        return False
+
+    def enter_shell(
+        self,
+        name: str,  # noqa: ARG002
+        path: str,
+        args: Iterable[str],
+    ):
         """
         Spawn a [shell](../../config/hatch.md#shell) within the environment.
 
@@ -694,7 +729,7 @@ class EnvironmentInterface(ABC):
         with self.command_context():
             self.platform.exit_with_command([path, *args])
 
-    def run_shell_command(self, command: str):
+    def run_shell_command(self, command: str, **kwargs):
         """
         This should return the standard library's
         [subprocess.CompletedProcess](https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess)
@@ -702,7 +737,8 @@ class EnvironmentInterface(ABC):
         [command_context](reference.md#hatch.env.plugin.interface.EnvironmentInterface.command_context)
         is active, with the expectation of providing the same guarantee.
         """
-        return self.platform.run_command(command, shell=True)
+        kwargs.setdefault('shell', True)
+        return self.platform.run_command(command, **kwargs)
 
     @contextmanager
     def command_context(self):
@@ -726,20 +762,20 @@ class EnvironmentInterface(ABC):
             yield from self.expand_command(command)
 
     def expand_command(self, command):
-        possible_script, args, ignore_exit_code = parse_script_command(command)
+        possible_script, args, _ignore_exit_code = parse_script_command(command)
 
         # Indicate undefined
         if not args:
             args = None
 
-        with self.metadata.context.apply_context(self.context):
+        with self.apply_context():
             if possible_script in self.scripts:
                 for cmd in self.scripts[possible_script]:
                     yield self.metadata.context.format(cmd, args=args).strip()
             else:
                 yield self.metadata.context.format(command, args=args).strip()
 
-    def construct_build_command(
+    def construct_build_command(  # noqa: PLR6301
         self,
         *,
         directory=None,
@@ -754,7 +790,7 @@ class EnvironmentInterface(ABC):
         This is the canonical way [`build`](../../cli/reference.md#hatch-build) command options are translated to
         a subprocess command issued to [builders](../builder/reference.md).
         """
-        command = ['python', '-u', '-m', 'hatchling', 'build', '--app']
+        command = ['python', '-u', '-m', 'hatchling', 'build']
 
         if directory:
             command.extend(('--directory', directory))
@@ -818,7 +854,8 @@ class EnvironmentInterface(ABC):
         and any method override should keep this check.
         """
         if self.platforms and self.platform.name not in self.platforms:
-            raise OSError('unsupported platform')
+            message = 'unsupported platform'
+            raise OSError(message)
 
     def get_env_vars(self) -> EnvVars:
         """
@@ -853,6 +890,11 @@ class EnvironmentInterface(ABC):
         """
         return {}
 
+    @contextmanager
+    def apply_context(self):
+        with self.get_env_vars(), self.metadata.context.apply_context(self.context):
+            yield
+
     def __enter__(self):
         self.activate()
         return self
@@ -864,11 +906,12 @@ class EnvironmentInterface(ABC):
 def expand_script_commands(env_name, script_name, commands, config, seen, active):
     if script_name in seen:
         return seen[script_name]
-    elif script_name in active:
+
+    if script_name in active:
         active.append(script_name)
-        raise ValueError(
-            f'Circular expansion detected for field `tool.hatch.envs.{env_name}.scripts`: {" -> ".join(active)}'
-        )
+
+        message = f'Circular expansion detected for field `tool.hatch.envs.{env_name}.scripts`: {" -> ".join(active)}'
+        raise ValueError(message)
 
     active.append(script_name)
 
